package com.qmth.exam.reserve.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.qmth.boot.core.collection.PageResult;
import com.qmth.boot.core.concurrent.service.ConcurrentService;
import com.qmth.boot.core.exception.StatusException;
import com.qmth.boot.tools.excel.ExcelReader;
import com.qmth.boot.tools.excel.enums.ExcelType;
import com.qmth.boot.tools.excel.model.DataMap;
import com.qmth.exam.reserve.bean.login.LoginUser;
import com.qmth.exam.reserve.bean.room.ExamRoomReq;
import com.qmth.exam.reserve.bean.room.ExamRoomSaveReq;
import com.qmth.exam.reserve.bean.room.ExamRoomVO;
import com.qmth.exam.reserve.cache.CacheConstants;
import com.qmth.exam.reserve.cache.impl.ApplyTaskCacheService;
import com.qmth.exam.reserve.dao.ExamRoomDao;
import com.qmth.exam.reserve.entity.ExamRoomEntity;
import com.qmth.exam.reserve.entity.ExamSiteEntity;
import com.qmth.exam.reserve.entity.TimePeriodEntity;
import com.qmth.exam.reserve.entity.TimePeriodExamRoomEntity;
import com.qmth.exam.reserve.service.*;
import com.qmth.exam.reserve.util.PageUtil;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.redisson.api.RLock;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.transaction.interceptor.TransactionAspectSupport;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

@Service
public class ExamRoomServiceImpl extends ServiceImpl<ExamRoomDao, ExamRoomEntity> implements ExamRoomService {

    private static final Logger log = LoggerFactory.getLogger(ExamRoomServiceImpl.class);
    private static final String[] EXCEL_HEADER = new String[]{"考场号", "考场名称", "考场地址", "考场容量", "所属考点代码", "所属考点名称"};

    @Autowired
    private ExamSiteService examSiteService;

    @Autowired
    private CategoryService categoryService;

    @Autowired
    private ApplyTaskCacheService cacheService;

    @Autowired
    private TimePeriodService timePeriodService;

    @Autowired
    private TimePeriodExamRoomService timePeriodExamRoomService;

    @Autowired
    private ConcurrentService concurrentService;


    @Override
    public PageResult<ExamRoomVO> pageExamRoom(ExamRoomReq req) {
        IPage<ExamRoomVO> iPage = baseMapper.pageExamRoom(new Page<>(req.getPageNumber(), req.getPageSize()),
                req);
        return PageUtil.of(iPage);
    }

    @Transactional
    @Override
    public void saveExamRoom(LoginUser user, ExamRoomSaveReq req) {
        if (concurrentService.isLocked(CacheConstants.LOCK_AUTO_APPLY)) {
            log.warn("[考场保存]系统自动预约中，不允许修改考场！lockKey:{}", CacheConstants.LOCK_AUTO_APPLY);
            throw new StatusException("系统正在自动预约中，不允许修改考场");
        }

        checkExamRoom(req);
        ExamRoomEntity examRoomEntity = new ExamRoomEntity();
        BeanUtils.copyProperties(req, examRoomEntity);
        ExamRoomEntity beforeUpdateRoom = new ExamRoomEntity();
        //考点容量
        ExamSiteEntity examSite = examSiteService.getById(req.getExamSiteId());
        int oldCapacity = examSite.getCapacity();
        int newCapacity = 0;
        if (req.getId() == null) {
            examRoomEntity.setEnable(Boolean.TRUE);
            save(examRoomEntity);
        } else {
            beforeUpdateRoom = getById(req.getId());
            updateById(examRoomEntity);
//            refreshApplyCountCache(examRoomEntity.getId(), examRoomEntity.getExamSiteId(), beforeUpdateRoom.getCapacity(), examRoomEntity.getCapacity());
        }
        //更新考点容量和教学点容量
        updateExamSiteAndTeachingCapacity(req.getExamSiteId());
        //考场更换考点
        if(req.getId() != null && !req.getExamSiteId().equals(beforeUpdateRoom.getExamSiteId())) {
            updateExamSiteAndTeachingCapacity(beforeUpdateRoom.getExamSiteId());
        }

        examSite = examSiteService.getById(req.getExamSiteId());
        newCapacity = examSite.getCapacity();

        //刷新缓存
        refreshApplyCountCache(examRoomEntity.getId(), examRoomEntity.getExamSiteId(), oldCapacity, newCapacity);
    }



    @Transactional
    @Override
    public void enable(Long id, Boolean enable) {
        if (concurrentService.isLocked(CacheConstants.LOCK_AUTO_APPLY)) {
            log.warn("[考场禁用/启用]系统自动预约中，不允许修改考场！lockKey:{}", CacheConstants.LOCK_AUTO_APPLY);
            throw new StatusException("系统正在自动预约中，不允许启用/禁用考场");
        }

        ExamRoomEntity examRoom = getById(id);
        //考点容量
        ExamSiteEntity examSite = examSiteService.getById(examRoom.getExamSiteId());
        int oldCapacity = examSite.getCapacity();

        LambdaUpdateWrapper<ExamRoomEntity> wrapper = new LambdaUpdateWrapper<>();
        wrapper.set(ExamRoomEntity::getEnable, enable);
        wrapper.eq(ExamRoomEntity::getId, id);
        update(null, wrapper);

        updateExamSiteAndTeachingCapacity(examRoom.getExamSiteId());

        //更新之后考点的容量
        examSite = examSiteService.getById(examRoom.getExamSiteId());

        //刷新缓存
        refreshApplyCountCache(id, examRoom.getExamSiteId(), oldCapacity, examSite.getCapacity());
    }

    @Override
    public List<Map<String, Object>> importExamRoom(LoginUser user, InputStream inputStream) {
        if (concurrentService.isLocked(CacheConstants.LOCK_AUTO_APPLY)) {
            log.warn("[考场导入]系统自动预约中，不允许修改考场！lockKey:{}", CacheConstants.LOCK_AUTO_APPLY);
            throw new StatusException("系统正在自动预约中，不允许导入考场");
        }

        List<DataMap> lineList;
        ExcelReader reader = ExcelReader.create(ExcelType.XLSX, inputStream, 0);
        try {
            lineList = reader.getDataMapList();
        } catch (Exception e) {
            throw new StatusException("Excel 解析失败");
        }
        if (com.baomidou.mybatisplus.core.toolkit.CollectionUtils.isEmpty(lineList)) {
            throw new StatusException("Excel无内容");
        }
        List<Map<String, Object>> failRecords = new ArrayList<>();
        List<ExamRoomEntity> examRoomList = new ArrayList<>();
        Long teachingId = user.getCategoryId();
        for (int i = 0; i < lineList.size(); i++) {
            DataMap line = lineList.get(i);
            ExamRoomEntity room = new ExamRoomEntity();
            StringBuilder msg = new StringBuilder();
            String code = trimAndNullIfBlank(line.get(EXCEL_HEADER[0]));
            if (StringUtils.isBlank(code)) {
                msg.append(" 考场号不能为空");
            } else if(code.length() > 20) {
                msg.append(" 考场号不能超过20个字符");
            } else {
                room.setCode(code);
            }
            String name = trimAndNullIfBlank(line.get(EXCEL_HEADER[1]));
            if (StringUtils.isBlank(name)) {
                msg.append(" 考场名称不能为空");
            } else if (name.length() > 50) {
                msg.append(" 考场名称不能超过50个字符");
            } else {
                room.setName(name);
            }
            String address = trimAndNullIfBlank(line.get(EXCEL_HEADER[2]));
            if (StringUtils.isBlank(address)) {
                msg.append(" 考场地址不能为空");
            } else if (address.length() > 100) {
                msg.append(" 考场地址不能超过100个字符");
            } else {
                room.setAddress(address);
            }
            String capacity = trimAndNullIfBlank(line.get(EXCEL_HEADER[3]));
            if (StringUtils.isBlank(capacity) || !capacity.matches("\\d+") || Integer.parseInt(capacity) <= 0 || Integer.parseInt(capacity) > 99999) {
                msg.append(" 考场容量必须为大于0且小于100000的整数");
            } else {
                room.setCapacity(Integer.parseInt(capacity));
            }
            String examSiteCode = trimAndNullIfBlank(line.get(EXCEL_HEADER[4]));
            if (StringUtils.isBlank(examSiteCode)) {
                msg.append(" 所属考点代码不能为空");
            }
            String examSiteName = trimAndNullIfBlank(line.get(EXCEL_HEADER[5]));
            if (StringUtils.isBlank(examSiteName)) {
                msg.append(" 所属考点名称不能为空");
            }
            List<ExamSiteEntity> examSiteList = listExamSite(examSiteCode, examSiteName, teachingId);
            if (examSiteList.isEmpty()) {
                msg.append(" 找不到所属考点");
            }
            if (examSiteList.size() > 1) {
                msg.append(" 找到的所属考点大于1个");
            }
            if (examSiteList.size() == 1) {
//                ExamRoomEntity examRoom = getExamRoom(examSiteList.get(0).getId(), code);
//                if (examRoom != null) {
//                    msg.append(" 当前考点下已经存在考场代码为：").append(code).append("的考场");
//                } else {
                    room.setExamSiteId(examSiteList.get(0).getId());
                    room.setEnable(Boolean.TRUE);
//                }
            }

            if (msg.length() > 0) {
                failRecords.add(newError(i + 1, msg.toString()));
            } else {
                examRoomList.add(room);
            }
        }

        if (!failRecords.isEmpty())
            return failRecords;

        for (int i = 0; i < examRoomList.size(); i++) {
            ExamRoomEntity examRoomEntity = examRoomList.get(i);
            try {
                saveRoom(examRoomEntity);
            } catch (Exception e) {
                failRecords.add(newError(i + 1, " 系统异常"));
                log.error("导入异常", e);
            }
        }

        if (CollectionUtils.isNotEmpty(failRecords)) {
            TransactionAspectSupport.currentTransactionStatus().setRollbackOnly();
            return failRecords;
        }

        return failRecords;
    }

    @Override
    public List<ExamRoomEntity> listExamRoom(Long examSiteId) {
        LambdaQueryWrapper<ExamRoomEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(ExamRoomEntity::getExamSiteId, examSiteId);
        wrapper.eq(ExamRoomEntity::getEnable, Boolean.TRUE);
        wrapper.orderByAsc(ExamRoomEntity::getCode);
        return list(wrapper);
    }

    @Override
    public List<ExamRoomEntity> listExamRoomByExamSiteId(Long examSiteId) {
        return baseMapper.listExamRoomByExamSiteId(examSiteId, Boolean.TRUE);
    }

    private void saveRoom(ExamRoomEntity room) {
        ExamRoomEntity examRoom = getExamRoom(room.getExamSiteId(), room.getCode());
        int oldCapacity = 0;
        ExamSiteEntity examSite;
        if (examRoom != null) {
            examSite = examSiteService.getById(examRoom.getExamSiteId());
            if(examSite == null) {
                throw new StatusException("找不到考点");
            }
            oldCapacity = examSite.getCapacity();
            examRoom.setCapacity(room.getCapacity());
            LambdaUpdateWrapper<ExamRoomEntity> wrapper = new LambdaUpdateWrapper<>();
            wrapper.set(ExamRoomEntity::getName, room.getName());
            wrapper.set(ExamRoomEntity::getAddress, room.getAddress());
            wrapper.set(ExamRoomEntity::getCapacity, room.getCapacity());
            wrapper.set(ExamRoomEntity::getEnable, Boolean.TRUE);
            wrapper.eq(ExamRoomEntity::getId, examRoom.getId());
            update(null, wrapper);
        } else {
            save(room);
        }
        //更新考点、教学点容量
        updateExamSiteAndTeachingCapacity(room.getExamSiteId());
        //更新之后的容量
        examSite = examSiteService.getById(room.getExamSiteId());
        //刷新缓存
        refreshApplyCountCache(room.getId(), room.getExamSiteId(), oldCapacity, examSite.getCapacity());
    }

    private List<ExamSiteEntity> listExamSite(String examSiteCode, String examSiteName, Long categoryId) {
        LambdaQueryWrapper<ExamSiteEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(ExamSiteEntity::getCode, examSiteCode);
        wrapper.eq(ExamSiteEntity::getName, examSiteName);
        wrapper.eq(categoryId != null, ExamSiteEntity::getCategoryId, categoryId);
        return examSiteService.list(wrapper);
    }

    private void updateExamSiteAndTeachingCapacity(Long examSiteId) {
        // 更新考点容量
        examSiteService.updateExamSiteCapacity(examSiteId);
        ExamSiteEntity newExamSite = examSiteService.getById(examSiteId);
        //更新教学点容量
        categoryService.updateTeachingCapacity(newExamSite.getCategoryId());
    }

    private void refreshApplyCountCache(Long examRoomId,Long examSiteId,  int oldCapacity, int newCapacity) {
        List<TimePeriodEntity> timePeriods = timePeriodService.list(new LambdaQueryWrapper<TimePeriodEntity>().select(TimePeriodEntity::getId));
        //考场时段列表
        List<TimePeriodExamRoomEntity> timePeriodExamRooms = timePeriodExamRoomService.listExamRoom(examRoomId);
        //将timePeriodExamRooms转Map
        Map<Long, TimePeriodExamRoomEntity> timePeriodExamRoomMap = timePeriodExamRooms.stream().collect(
                Collectors.toMap(TimePeriodExamRoomEntity::getTimePeriodId, Function.identity()));

        for (TimePeriodEntity timePeriod : timePeriods) {
            String examSiteTimePeriodLockKey = String.format(CacheConstants.LOCK_EXAM_SITE_TIME_PERIOD, examSiteId, timePeriod.getId());
            RLock examSiteTimePeriodLock = (RLock) concurrentService.getLock(examSiteTimePeriodLockKey);
            try {
                if (!examSiteTimePeriodLock.tryLock()) {
                    log.warn("[考场容量更新] 获取锁失败，不能同时修改考点考场容量, lockKey:{}", examSiteTimePeriodLock);
                    throw new StatusException("考场保存失败，请稍后再试");
                } else {
                    //刷新缓存
                    TimePeriodExamRoomEntity timePeriodExamRoomEntity = timePeriodExamRoomMap.get(timePeriod.getId());
                    if (timePeriodExamRoomEntity == null || timePeriodExamRoomEntity.getEnable()) {
                        cacheService.refreshApplyAvailableCountCache(examSiteId, timePeriod.getId(), oldCapacity, newCapacity);
                    }
                }
            } catch (Exception e) {
                log.error("[考场容量更新] 考场保存失败, msg:{}", e.getMessage());
                throw new StatusException("考场保存失败，请稍后再试");
            } finally {
                try {
                    if (examSiteTimePeriodLock.isLocked() && examSiteTimePeriodLock.isHeldByCurrentThread()) {
                        examSiteTimePeriodLock.unlock();
                        log.info("[考场容量更新] 解锁成功,lockKey:{}", examSiteTimePeriodLock);
                    }
                } catch (Exception e) {
                    log.warn(e.getMessage());
                }
            }

        }
    }

    private void checkExamRoom(ExamRoomSaveReq req) {
        if (StringUtils.isEmpty(req.getCode())) {
            throw new StatusException("考场代码不能为空");
        }
        if (StringUtils.isEmpty(req.getName())) {
            throw new StatusException("考场名称不能为空");
        }
        if (StringUtils.isEmpty(req.getAddress())) {
            throw new StatusException("考场地址不能为空");
        }
        if (req.getCapacity() == null || req.getCapacity() <= 0) {
            throw new StatusException("考场容量必须大于0");
        }
        if (req.getExamSiteId() == null) {
            throw new StatusException("请选择考点");
        }
        checkExamRoomExist(req, req.getId() != null);
    }

    private void checkExamRoomExist(ExamRoomSaveReq req, boolean idFlag) {
        LambdaQueryWrapper<ExamRoomEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(ExamRoomEntity::getExamSiteId, req.getExamSiteId());
        wrapper.eq(ExamRoomEntity::getCode, req.getCode());
        wrapper.notIn(idFlag, ExamRoomEntity::getId, req.getId());
        ExamRoomEntity examRoomEntity = baseMapper.selectOne(wrapper);
        if (examRoomEntity != null) {
            ExamSiteEntity site = examSiteService.getById(req.getExamSiteId());
            throw new StatusException(site.getName() + "下已存在代码为:" + req.getCode() + "的考场");
        }
    }

    private ExamRoomEntity getExamRoom(Long examSiteId, String roomCode) {
        LambdaQueryWrapper<ExamRoomEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(ExamRoomEntity::getExamSiteId, examSiteId);
        wrapper.eq(ExamRoomEntity::getCode, roomCode);
        return baseMapper.selectOne(wrapper);
    }

    private String trimAndNullIfBlank(String s) {
        if (StringUtils.isBlank(s)) {
            return null;
        }
        return s.trim();
    }

    private Map<String, Object> newError(int lineNum, String msg) {
        Map<String, Object> map = new HashMap<>();
        map.put("lineNum", lineNum);
        map.put("msg", msg);
        return map;
    }
}
